//
//  SheetContainerViewController.swift
//  SheetImagePicker
//
//  Created by Shadowfacts on 9/23/19.
//  Copyright © 2019 Shadowfacts. All rights reserved.
//

import UIKit

public protocol SheetContainerViewControllerDelegate {
    func sheetContainer(_ sheetContainer: SheetContainerViewController, willSnapToDetent detent: Detent) -> Bool
    func sheetContainer(_ sheetContainer: SheetContainerViewController, didSnapToDetent detent: Detent)
}

// default no-op implementation
public extension SheetContainerViewControllerDelegate {
    func sheetContainer(_ sheetContainer: SheetContainerViewController, willSnapToDetent detent: Detent) -> Bool {
        return true
    }
    func sheetContainer(_ sheetContainer: SheetContainerViewController, didSnapToDetent detent: Detent) {}
}

public class SheetContainerViewController: UIViewController {

    public var delegate: SheetContainerViewControllerDelegate?
    
    let content: UIViewController
    
    public var detents: [Detent] = [.bottom, .middle, .top]
    var topDetent: (detent: Detent, offset: CGFloat) {
        return detents.map { ($0, $0.offset(in: view)) }.min(by: { $0.1 < $1.1 })!
    }

    var topConstraint: NSLayoutConstraint!
    lazy var initialConstant: CGFloat = view.bounds.height / 2
    
    public var minimumDetentJumpVelocity: CGFloat = 500
    public var maximumStretchDistance: CGFloat = 15
    
    public init(content: UIViewController) {
        self.content = content
        
        super.init(nibName: nil, bundle: nil)
    }
    
    required public init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }

    override public func viewDidLoad() {
        super.viewDidLoad()
        
        addChild(content)
        content.didMove(toParent: self)
        view.addSubview(content.view)
        topConstraint = content.view.topAnchor.constraint(equalTo: view.topAnchor, constant: initialConstant)
        NSLayoutConstraint.activate([
            content.view.leadingAnchor.constraint(equalTo: view.leadingAnchor),
            content.view.trailingAnchor.constraint(equalTo: view.trailingAnchor),
            topConstraint,
            content.view.bottomAnchor.constraint(equalTo: view.bottomAnchor)
        ])
        
        let panGesture = UIPanGestureRecognizer(target: self, action: #selector(panGestureRecognized(_:)))
        content.view.addGestureRecognizer(panGesture)
    }

    @objc func panGestureRecognized(_ recognizer: UIPanGestureRecognizer) {
        switch recognizer.state {
        case .began:
            initialConstant = topConstraint.constant
            
        case .changed:
            let translation = recognizer.translation(in: content.view)
            var realOffset = initialConstant + translation.y
            if realOffset < topDetent.offset {
                func clamp(_ value: CGFloat, from: CGFloat, to: CGFloat) -> CGFloat {
                    if value < from {
                        return from
                    } else if value > to {
                        return to
                    } else {
                        return value
                    }
                }
                func smoothstep(value: CGFloat, from: CGFloat, to: CGFloat) -> CGFloat {
                    let x = clamp((value - from) / (to - from), from: 0, to: 1)
                    // 3x^2 - 2x^3
                    return 3 * pow(x, 2) - 2 * pow(x, 3)
                }
                let topOffset = topDetent.offset
                let smoothed = smoothstep(value: realOffset, from: topOffset, to: 0)
                realOffset = topOffset - smoothed * maximumStretchDistance
                
            }
            topConstraint.constant = realOffset
            
        case .ended:
            let velocity = recognizer.velocity(in: view)

            let springToDetent: (Detent, CGFloat)
            if abs(velocity.y) > minimumDetentJumpVelocity,
                let detentInVelocityDirection = nearestDetentOffset(currentOffset: topConstraint.constant, direction: velocity.y) {
                springToDetent = detentInVelocityDirection
            } else if let nearestDetent = nearestDetentOffset(currentOffset: topConstraint.constant) {
                springToDetent = nearestDetent
            } else {
                return
            }
            
            if delegate?.sheetContainer(self, willSnapToDetent: springToDetent.0) ?? true {
                let springDistance = abs(topConstraint.constant - springToDetent.1)
                self.topConstraint.constant = springToDetent.1
                let springVelocity = velocity.y / springDistance
                
                UIView.animate(withDuration: 0.35, delay: 0, usingSpringWithDamping: 0.6, initialSpringVelocity: springVelocity, animations: {
                    self.view.layoutIfNeeded()
                }, completion: { (finished) in
                    self.delegate?.sheetContainer(self, didSnapToDetent: springToDetent.0)
                })
            }
        default:
            return
        }
    }
    
    func nearestDetentOffset(currentOffset: CGFloat) -> (Detent, CGFloat)? {
        return detents.map { ($0, $0.offset(in: view)) }.min { (a, b) -> Bool in
            return abs(a.1 - currentOffset) < abs(b.1 - currentOffset)
        }
    }
    
    func nearestDetentOffset(currentOffset: CGFloat, direction: CGFloat) -> (Detent, CGFloat)? {
        let sorted = detents.map { ($0, $0.offset(in: view)) }.sorted { (a, b) -> Bool in
            return a.1 < b.1
        }
        if direction < 0 {
            return sorted.last(where: { $0.1 < currentOffset })
        } else {
            return sorted.first(where: { $0.1 > currentOffset })
        }
    }
    
}

